/***************************************************************************
                          thread.h  -  description
                             -------------------
    begin                : Thu Jan 6 2000
    copyright            : (C) 2000 by Gavin James Wood
    email                : gav@indigoarchive.net
 ***************************************************************************/

#ifndef _SUCS_THREAD
#define _SUCS_THREAD

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#pragma interface

#include <map>
#include <string>
#include <iostream>
using namespace std;

#include <pthread.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#ifdef __SUCS
#include "mutex.h"
#include "sema.h"
#else
#include <sucs/mutex.h>
#include <sucs/sema.h>
#endif
using namespace sucs;

namespace sucs
{

extern Sema __output;

#define terr(X) { __output.wait(); cerr << "[" << (*Thread::me()) << "]: " << X << endl << flush; __output.signal(); }
#ifdef _SUCS_DEBUG_MESSAGES
#define tdbg(X) terr(X)
#else
#define tdbg(X)
#endif

class Thread
{
	Sema theDead, theStartingUp;
	Mutex theJoining;
	
	friend void *gameOn(void *arg);
	friend void cleanup(void *arg);
	
	mutable bool theRunning, theJoinable;
	string theName;
	int theId;

	mutable pthread_t theThreadId;
	static map<pthread_t, Thread *> theThreads;
	static int theIdCount, theCount, theNowRunningCount;
	static Sema theNowRunning;
	static pthread_t theMainId;

protected:
	/** The over-loadable function where the thread code goes. */
	virtual void body() = 0;
	
	/** The over-loadable function called when the thread is all but dead.
	 * Does nothing by default.
	 */
	virtual void ghost() { }

public:
	/** Checks whether this thread is running or not. */
	bool isRunning() const;

	/** Stop the current thread, waits for it to exit, then starts a new thread. */
	bool safelyStart(bool joinable = false, const unsigned int stackSize = 16384);

	/** Starts a new thread or fails if the thread is already active. */
	bool start(bool joinable = false, const unsigned int stackSize = 16384);

	/** Issues a cancelation request to the thread. Asynchronous. */
	bool cancel() const;
	
	/** Cancels the thread and doesn't return till dead.
	 * Should be used judiciously in Thread-derived destructor functions.
	 */
	bool cancelAndWait() const;
	
	/** Get the automatically assigned id of the Thread (starts at 1, with 0 == main()'s thread ). */
	const int id() const { return theId; }

	/** Get the pthread id of the thread. */
	const pthread_t threadId() const { return theThreadId; }

	/** Get the given name of this thread. */
	const string &name() const { return theName; }

	/** Nice ostream operator. */
	friend ostream &operator<<(ostream &out, const Thread &t);
	
	/** Will block until the thread is no longer running. */
	void waitForDeath() const;
	
	/** Will join with the thread (i.e. block until it's completed and free its resources).
	 * 
	 * @return Returns void * which does nothing currently.
	 */
	void *join() const;

	// *** STATIC METHODS
	
	/** Use to get pthread identifier of caller. */
	inline static pthread_t myThreadId() { return pthread_self(); }
	
	/** Use to get associated Thread object for the calling thread. Returns NULL in case of main thread. */
	static Thread *me() { if(theThreads.count(pthread_self())) return theThreads[pthread_self()]; else return NULL; }
	
	/** Use to get numeric ID of calling thread (0 for main thread). */
	static int myId() { if(me()) return me()->id(); else return 0; }

	/** Use to get name of the calling thread ("main" for main thread). */
	static string myName() { if(me()) return me()->name(); else return "main"; }

	/**
	 * Will allow thread to be cancelled *immediately* rather than at the next cancellation point.
	 */
	static void setMyImmediateCancel(bool immediate);
	
	// *** CONSTRUCTORS/DESTRUCTORS
	
	Thread(const string &name = "");
	virtual ~Thread();
};

}

#endif
